@@ -6,7 +6,7 @@
  */
 
 /*
- * Copyright (c) 2006-2011 Logitech.
+ * Copyright (c) 2006-2009 Logitech.
  *
  * This file is part of libwebcam.
  * 
@@ -30,17 +30,37 @@
 #include <linux/videodev2.h>
 
 #ifdef USE_UVCVIDEO
-	#ifndef USE_PRIVATE_UVCVIDEO_H
-		// By default use the public uvcvideo.h include file
-		#include <linux/uvcvideo.h>
-	#else
-		// For pre-2.6.38 use the header file from the uvcvideo sources
-		#define __user
-		#include <uvcvideo.h>
-	#endif
+#define __user
+#include <linux/uvcvideo.h>
 #endif
 
+#ifndef UVC_CONTROL_SET_CUR
+/* Control flags */
+#define UVC_CONTROL_SET_CUR	(1 << 0)
+#define UVC_CONTROL_GET_CUR	(1 << 1)
+#define UVC_CONTROL_GET_MIN	(1 << 2)
+#define UVC_CONTROL_GET_MAX	(1 << 3)
+#define UVC_CONTROL_GET_RES	(1 << 4)
+#define UVC_CONTROL_GET_DEF	(1 << 5)
+/* Control should be saved at suspend and restored at resume. */
+#define UVC_CONTROL_RESTORE	(1 << 6)
+/* Control can be updated by the camera. */
+#define UVC_CONTROL_AUTO_UPDATE	(1 << 7)
 
+#define UVC_CONTROL_GET_RANGE	(UVC_CONTROL_GET_CUR | UVC_CONTROL_GET_MIN | \
+				 UVC_CONTROL_GET_MAX | UVC_CONTROL_GET_RES | \
+				 UVC_CONTROL_GET_DEF)
+#endif
+
+#ifndef UVCIOC_CTRL_ADD
+struct uvc_xu_control_info {
+	__u8 entity[16];
+	__u8 index;
+	__u8 selector;
+	__u16 size;
+	__u32 flags;
+};
+#endif
 // uvcvideo.h
 
 // Old uvcvideo pre-r178 control IDs
@@ -60,40 +80,6 @@
 #define UVC_CID_WHITE_BALANCE_TEMPERATURE_AUTO	(V4L2_CID_PRIVATE_BASE+12)
 #define UVC_CID_WHITE_BALANCE_TEMPERATURE		(V4L2_CID_PRIVATE_BASE+13)
 
-// UVC control flags (pre-2.6.38)
-#ifndef UVC_CTRL_FLAG_SET_CUR
-#define UVC_CTRL_FLAG_SET_CUR UVC_CONTROL_SET_CUR
-#endif
-#ifndef UVC_CTRL_FLAG_GET_CUR
-#define UVC_CTRL_FLAG_GET_CUR UVC_CONTROL_GET_CUR
-#endif
-#ifndef UVC_CTRL_FLAG_GET_MIN
-#define UVC_CTRL_FLAG_GET_MIN UVC_CONTROL_GET_MIN
-#endif
-#ifndef UVC_CTRL_FLAG_GET_MAX
-#define UVC_CTRL_FLAG_GET_MAX UVC_CONTROL_GET_MAX
-#endif
-#ifndef UVC_CTRL_FLAG_GET_RES
-#define UVC_CTRL_FLAG_GET_RES UVC_CONTROL_GET_RES
-#endif
-#ifndef UVC_CTRL_FLAG_GET_DEF
-#define UVC_CTRL_FLAG_GET_DEF UVC_CONTROL_GET_DEF
-#endif
-#ifndef UVC_CTRL_FLAG_AUTO_UPDATE
-#define UVC_CTRL_FLAG_AUTO_UPDATE UVC_CONTROL_AUTO_UPDATE
-#endif
-
-// UVC ioctl types (pre-2.6.42)
-#ifndef UVCIOC_CTRL_ADD
-struct uvc_xu_control_info {
-	__u8 entity[16];
-	__u8 index;
-	__u8 selector;
-	__u16 size;
-	__u32 flags;
-};
-#endif
-
 
 // videodev2.h
 
