@@ -6,7 +6,7 @@
  */
 
 /*
- * Copyright (c) 2006-2011 Logitech.
+ * Copyright (c) 2006-2009 Logitech.
  *
  * This file is part of libwebcam.
  * 
@@ -35,6 +35,7 @@
 #include <unistd.h>
 #include <stdarg.h>
 #include <sys/ioctl.h>
+#include <linux/videodev2.h>
 #include <errno.h>
 #include <iconv.h>
 
@@ -138,6 +139,8 @@ typedef struct _UVCXUControl {
 	struct uvc_xu_control_info	info;
 	/// Pointer to the next extension unit control definition in the list
 	struct _UVCXUControl		* next;
+	/// Does the match section of this control match the current device
+	int				match;
 
 } UVCXUControl;
 
@@ -147,6 +150,8 @@ typedef struct _UVCXUControl {
 typedef struct _ParseContext {
 	/// Structure used to pass information between the application and libwebcam. Can be NULL.
 	CDynctrlInfo	* info;
+	/// XML document tree representing a dynamic controls configuration
+	xmlDoc		* xml_doc;
 	/// Size of the info->messages buffer (which contains the CDynctrlMessage structures
 	/// and the strings pointed to).
 	unsigned int	messages_size;
@@ -154,6 +159,8 @@ typedef struct _ParseContext {
 	iconv_t			cd;
 	/// List of constants parsed from the @c constants node
 	Constant		* constants;
+	/// Device info
+	CDevice			* device;
 	/// Handle to the libwebcam device
 	CHandle			handle;
 	/// Handle to the V4L2 device that is used to add the dynamic controls
@@ -336,31 +343,31 @@ static int string_to_version (const char *string, unsigned int *major, unsigned
  * 		- 0 if the conversion failed, i.e. the request name was not recognized
  * 		- a UVC_CONTROL_* constant if the conversion was successful
  */
-static __u32 get_uvc_request_by_name (const xmlChar *name)
-{
-	__u32 request = 0;	// Used to denote an invalid/unsupported UVC control request
-	if(!name) return request;
-
-	if(xmlStrEqual(name, BAD_CAST("SET_CUR"))) {
-		request = UVC_CTRL_FLAG_SET_CUR;
-	}
-	else if(xmlStrEqual(name, BAD_CAST("GET_CUR"))) {
-		request = UVC_CTRL_FLAG_GET_CUR;
-	}
-	else if(xmlStrEqual(name, BAD_CAST("GET_MIN"))) {
-		request = UVC_CTRL_FLAG_GET_MIN;
-	}
-	else if(xmlStrEqual(name, BAD_CAST("GET_MAX"))) {
-		request = UVC_CTRL_FLAG_GET_MAX;
-	}
-	else if(xmlStrEqual(name, BAD_CAST("GET_RES"))) {
-		request = UVC_CTRL_FLAG_GET_RES;
-	}
-	else if(xmlStrEqual(name, BAD_CAST("GET_DEF"))) {
-		request = UVC_CTRL_FLAG_GET_DEF;
-	}
-	return request;
-}
+//static __u32 get_uvc_request_by_name (const xmlChar *name)
+//{
+//	__u32 request = 0;	// Used to denote an invalid/unsupported UVC control request
+//	if(!name) return request;
+//
+//	if(xmlStrEqual(name, BAD_CAST("SET_CUR"))) {
+//		request = UVC_CONTROL_SET_CUR;
+//	}
+//	else if(xmlStrEqual(name, BAD_CAST("GET_CUR"))) {
+//		request = UVC_CONTROL_GET_CUR;
+//	}
+//	else if(xmlStrEqual(name, BAD_CAST("GET_MIN"))) {
+//		request = UVC_CONTROL_GET_MIN;
+//	}
+//	else if(xmlStrEqual(name, BAD_CAST("GET_MAX"))) {
+//		request = UVC_CONTROL_GET_MAX;
+//	}
+//	else if(xmlStrEqual(name, BAD_CAST("GET_RES"))) {
+//		request = UVC_CONTROL_GET_RES;
+//	}
+//	else if(xmlStrEqual(name, BAD_CAST("GET_DEF"))) {
+//		request = UVC_CONTROL_GET_DEF;
+//	}
+//	return request;
+//}
 
 
 /**
@@ -422,18 +429,18 @@ static enum v4l2_ctrl_type get_v4l2_ctrl_type_by_name (const xmlChar *name)
 	else if(xmlStrEqual(name, BAD_CAST("V4L2_CTRL_TYPE_BOOLEAN"))) {
 		type = V4L2_CTRL_TYPE_BOOLEAN;
 	}
+	else if(xmlStrEqual(name, BAD_CAST("V4L2_CTRL_TYPE_BUTTON"))) {
+		type = V4L2_CTRL_TYPE_BUTTON;
+	}
+	else if(xmlStrEqual(name, BAD_CAST("V4L2_CTRL_TYPE_MENU"))) {
+		type = V4L2_CTRL_TYPE_MENU;
+	}
 #ifdef ENABLE_RAW_CONTROLS
 	else if(xmlStrEqual(name, BAD_CAST("V4L2_CTRL_TYPE_STRING"))) {
 		type = V4L2_CTRL_TYPE_STRING;
 	}
 #endif
 	/*
-	else if(xmlStrEqual(name, BAD_CAST("V4L2_CTRL_TYPE_MENU"))) {
-		type = V4L2_CTRL_TYPE_MENU;
-	}
-	else if(xmlStrEqual(name, BAD_CAST("V4L2_CTRL_TYPE_BUTTON"))) {
-		type = V4L2_CTRL_TYPE_BUTTON;
-	}
 	else if(xmlStrEqual(name, BAD_CAST("V4L2_CTRL_TYPE_INTEGER64"))) {
 		type = V4L2_CTRL_TYPE_INTEGER64;
 	}
@@ -920,11 +927,7 @@ static CResult add_error_at_node (ParseContext *ctx, const xmlNode *node, const
 /**
  * Parse a dynamic controls configuration XML file and return an XML document tree.
  *
- * Note that the pointer stored in @a *xml_doc must be freed using xmlFreeDoc() if
- * the return value of this functino is C_SUCCESS.
- *
- * @param file_name		name (with an optional path) of the file to be parsed
- * @param xml_doc		address of a pointer to store the XML document tree
+ * @param file_name	name (with an optional path) of the file to be parsed
  * @param ctx		current parse context
  *
  * @return
@@ -932,20 +935,18 @@ static CResult add_error_at_node (ParseContext *ctx, const xmlNode *node, const
  * 		- C_PARSE_ERROR if the XML file is malformed
  * 		- C_SUCCESS if parsing was successful
  */
-static CResult parse_dynctrl_file (const char *file_name, xmlDoc **xml_doc, ParseContext *ctx)
+static CResult parse_dynctrl_file (const char *file_name, ParseContext *ctx)
 {
 	CResult ret = C_SUCCESS;
 	xmlParserCtxt *parser = NULL;
-	xmlDoc *doc = NULL;
-	assert(xml_doc);
 
 	parser = xmlNewParserCtxt();
 	if(!parser)
 		return C_NO_MEMORY;
 	
 	// Read and parse the XML file
-	doc = xmlCtxtReadFile(parser, file_name, NULL, XML_PARSE_NOBLANKS);
-	if(!doc) {
+	ctx->xml_doc = xmlCtxtReadFile(parser, file_name, NULL, XML_PARSE_NOBLANKS);
+	if(!ctx->xml_doc) {
 		xmlError *e = &parser->lastError;
 		add_message(ctx, e->line, e->int2, CD_SEVERITY_ERROR,
 				"Malformed control mapping file encountered. Unable to parse: %s",
@@ -960,13 +961,13 @@ static CResult parse_dynctrl_file (const char *file_name, xmlDoc **xml_doc, Pars
 	}
 
 	// Free the document tree if there was an error
-	if(ret) {
-		xmlFreeDoc(doc);
-		doc = NULL;
-	}
+	//if(ret) {
+	//	xmlFreeDoc(doc);
+	//	doc = NULL;
+	//}
 
 	// Return the document (or NULL if there was an error)
-	*xml_doc = doc;
+	//*xml_doc = doc;
 	
 	// Clean up
 	xmlFreeParserCtxt(parser);
@@ -1022,6 +1023,10 @@ static CResult process_mapping (const xmlNode *node_mapping, ParseContext *ctx)
 		return C_PARSE_ERROR;
 	}
 	xmlFree(control_ref); control_ref = NULL;
+	
+	if (!control->match)
+		return C_SUCCESS;
+	
 	memcpy(mapping_info.entity, control->info.entity, GUID_SIZE);
 	mapping_info.selector = control->info.selector;
 
@@ -1088,6 +1093,66 @@ static CResult process_mapping (const xmlNode *node_mapping, ParseContext *ctx)
 		return C_PARSE_ERROR;
 	}
 	mapping_info.data_type = uvc_type;
+	
+	if(v4l2_type == V4L2_CTRL_TYPE_MENU) {
+		xmlNode *node_menu = xml_get_first_child_by_name(node_v4l2, "menu_entry");
+		if(!node_menu) {
+			add_error_at_node(ctx, node_v4l2,
+				"<menu_entry> is mandatory for mappings with a v4l2_type of V4L2_CTRL_TYPE_MENU");
+			return C_PARSE_ERROR;
+		}
+
+		mapping_info.menu_count = 1;
+		while ((node_menu = xml_get_next_sibling_by_name(node_menu, "menu_entry")))
+			mapping_info.menu_count++;
+
+		mapping_info.menu_info =
+			malloc(mapping_info.menu_count * sizeof(struct uvc_menu_info));
+		if(!mapping_info.menu_info)
+			return C_NO_MEMORY;
+
+		int i;
+		node_menu = xml_get_first_child_by_name(node_v4l2, "menu_entry");
+		for (i = 0; i < mapping_info.menu_count; i++) {
+			xmlChar *menu_name_uni = xmlGetProp(node_menu, BAD_CAST("name"));
+			char *menu_name_asc = UNICODE_TO_NORM_ASCII(menu_name_uni);
+			if(!menu_name_asc) {
+				add_error_at_node(ctx, node_menu,
+					"Invalid menu_entry. 'name' attribute is mandatory.");
+				free(mapping_info.menu_info);
+				return C_PARSE_ERROR;
+			}
+
+			xmlChar *menu_value = xmlGetProp(node_menu, BAD_CAST("value"));
+			if(!menu_value) {
+				add_error_at_node(ctx, node_menu,
+					"Invalid menu_entry. 'value' attribute is mandatory.");
+				xmlFree(menu_name_uni);
+				free(menu_name_asc);
+				free(mapping_info.menu_info);
+				return C_PARSE_ERROR;
+			}
+
+			snprintf((char *)mapping_info.menu_info[i].name,
+				 sizeof(mapping_info.menu_info[i].name),
+				 "%s", menu_name_asc);
+			ret = lookup_or_convert_to_integer(menu_value,
+				(int *)&mapping_info.menu_info[i].value,
+				ctx);
+			if(ret)
+				add_error_at_node(ctx, node_menu,
+					"<menu_entry> value contains invalid number or references unknown constant: '%s'",
+					menu_value);
+			xmlFree(menu_value);
+			xmlFree(menu_name_uni);
+			free(menu_name_asc);
+			if(ret) {
+				free(mapping_info.menu_info);
+				return C_PARSE_ERROR;
+			}
+			node_menu = xml_get_next_sibling_by_name(node_menu, "menu_entry");
+		}
+	}
 
 	// Add the control to the UVC driver's control list
 	/*
@@ -1113,6 +1178,7 @@ static CResult process_mapping (const xmlNode *node_mapping, ParseContext *ctx)
 	);
 	*/
 	int v4l2_ret = ioctl(ctx->v4l2_handle, UVCIOC_CTRL_MAP, &mapping_info);
+	free(mapping_info.menu_info);
 	if(v4l2_ret != 0
 #ifdef DYNCTRL_IGNORE_EEXIST_AFTER_PASS1
 			&& (ctx->pass == 1 || errno != EEXIST)
@@ -1157,7 +1223,7 @@ static CResult process_mappings (const xmlNode *node_mappings, ParseContext *ctx
 /**
  * Process a @c control node by adding the contained control to the UVC driver.
  */
-static CResult process_control (xmlNode *node_control, ParseContext *ctx)
+static CResult process_control (xmlNode *node_control, ParseContext *ctx, int match)
 {
 	CResult ret = C_SUCCESS;
 	assert(node_control);
@@ -1171,6 +1237,9 @@ static CResult process_control (xmlNode *node_control, ParseContext *ctx)
 		return C_NO_MEMORY;
 	memset(xu_control, 0, sizeof *xu_control);
 
+	// Set match
+	xu_control->match = match;
+
 	// Get the ID of the extension unit control definition
 	xu_control->id = xmlGetProp(node_control, BAD_CAST("id"));
 	if(!xu_control->id) {
@@ -1201,89 +1270,8 @@ static CResult process_control (xmlNode *node_control, ParseContext *ctx)
 	}
 	xu_control->info.selector = (__u8)value;
 
-	// Retrieve the index
-	text = xml_get_node_text(xml_get_first_child_by_name(node_control, "index"));
-	ret = lookup_or_convert_to_integer(text, &value, ctx);
-	if(ret) {
-		add_error_at_node(ctx, node_control,
-			"Invalid control index specified: '%s'", text ? (char *)text : "<empty>");
-		goto done;
-	}
-	xu_control->info.index = (__u8)value;
-
-	// Retrieve the size
-	text = xml_get_node_text(xml_get_first_child_by_name(node_control, "size"));
-	ret = lookup_or_convert_to_integer(text, &value, ctx);
-	if(ret || !is_valid_size(value, 0xFFFF)) {
-		add_error_at_node(ctx, node_control,
-			"Invalid control size specified: '%s'", text ? (char *)text : "<empty>");
-		goto done;
-	}
-	xu_control->info.size = (__u16)value;
-
-	// Retrieve the list of supported requests
-	xmlNode *node_requests = xml_get_first_child_by_name(node_control, "requests");
-	if(!node_requests) {
-		add_error_at_node(ctx, node_control,
-			"List of supported UVC requests missing. <requests> is mandatory.");
-		ret = C_PARSE_ERROR;
-		goto done;
-	}
-	xmlNode *node_request = xml_get_first_child_by_name(node_requests, "request");
-	while(node_request) {
-		text = xml_get_node_text(node_request);
-		__u32 flag = get_uvc_request_by_name(text);
-		if(!flag) {
-			add_error_at_node(ctx, node_request,
-				"Invalid UVC request specified: '%s'", text ? (char *)text : "<empty>");
-		}
-		xu_control->info.flags |= flag;
-		node_request = xml_get_next_sibling_by_name(node_request, "request");
-	}
-
-	// Disable the UVC driver's caching mechanism for XU controls
-	xu_control->info.flags |= UVC_CTRL_FLAG_AUTO_UPDATE;
-
-	// Add the control to the UVC driver's control list
-	/*
-	printf(
-		"uvc_xu_control_info = {\n"
-		"	entity   = {" GUID_FORMAT "},\n"
-		"	index    = %u,\n"
-		"	selector = %u,\n"
-		"	size     = %u,\n"
-		"	flags    = %u\n"
-		"}\n",
-		GUID_ARGS(xu_control->info.entity),
-		xu_control->info.index,
-		xu_control->info.selector,
-		xu_control->info.size,
-		xu_control->info.flags
-	);
-	*/
-#ifdef UVCIOC_CTRL_ADD
-	int v4l2_ret = ioctl(ctx->v4l2_handle, UVCIOC_CTRL_ADD, &xu_control->info);
-	if(v4l2_ret != 0
-#ifdef DYNCTRL_IGNORE_EEXIST_AFTER_PASS1
-			&& (ctx->pass == 1 || errno != EEXIST)
-#endif
-		)
-	{
-		add_error(ctx,
-			"%s: unable to add control with GUID {"GUID_FORMAT"} and selector %d. "
-			"ioctl(UVCIOC_CTRL_ADD) failed with return value %d (error %d: %s)",
-			GET_HANDLE(ctx->handle).device->v4l2_name,
-			GUID_ARGS(xu_control->info.entity), xu_control->info.selector,
-			v4l2_ret, errno, strerror(errno));
-		ret = C_V4L2_ERROR;
-	}
-#endif
-
 	// Add the extension unit control definition to the internal list for later
 	// reference by mappings.
-	// Note that we add controls to the list whether or not the UVCIOC_CTRL_ADD above
-	// succeeded or not. The control might already exist in the driver but a referencing
-	// mapping might not, but the mapping needs to be able to look up the control.
 	xu_control->next = ctx->controls;
 	ctx->controls = xu_control;
 
@@ -1296,18 +1284,17 @@ done:
 	return ret;
 }
 
-
 /**
  * Process a @c controls node.
  */
-static CResult process_controls (const xmlNode *node_controls, ParseContext *ctx)
+static CResult process_controls (const xmlNode *node_controls, ParseContext *ctx, int match)
 {
 	assert(node_controls);
 
 	// Process all <control> nodes
 	xmlNode *node_control = xml_get_first_child_by_name(node_controls, "control");
 	while(node_control) {
-		CResult ret = process_control(node_control, ctx);
+		CResult ret = process_control(node_control, ctx, match);
 		if(ctx->info) {
 			if(ret)
 				ctx->info->stats.controls.successful++;
@@ -1328,12 +1315,54 @@ static CResult process_controls (const xmlNode *node_controls, ParseContext *ctx
  */
 static CResult process_device (const xmlNode *node_device, ParseContext *ctx)
 {
+	xmlNode *node_match;
+	
 	assert(node_device);
+	int match = 1; /* An entry without match sections always matches */
 
+	// Process match
+	for (node_match = xml_get_first_child_by_name(node_device, "match");
+	     node_match != NULL;
+	     node_match = xml_get_next_sibling_by_name(node_match, "match")) {
+		int vid, pid;
+
+		match = 0;
+
+		xmlNode *node = xml_get_first_child_by_name(node_match, "vendor_id");
+		if (node == NULL ||
+		    !is_valid_integer_string((char *)xml_get_node_text(node), &vid)) {
+			add_error_at_node(ctx, node_match,
+				"<vendor_id> is mandatory for match sections and must be a valid integer");
+			return C_PARSE_ERROR;
+		}
+		if (vid != ctx->device->usb.vendor)
+			continue;
+
+		match = 1; /* Sessions with no product_id match on vendor only */
+
+		for (node = xml_get_first_child_by_name(node_match, "product_id");
+		     node != NULL;
+		     node = xml_get_next_sibling_by_name(node, "product_id")) {
+			match = 0;
+
+			if (!is_valid_integer_string((char *)xml_get_node_text(node), &pid)) {
+				add_error_at_node(ctx, node_match,
+					"<product_id> must be a valid integer");
+				return C_PARSE_ERROR;
+			}
+
+			if (pid == ctx->device->usb.product) {
+				match = 1;
+				break;
+			}
+		}
+		break;
+	}
+	
 	// Process the <controls> node
 	xmlNode *node_controls = xml_get_first_child_by_name(node_device, "controls");
 	if(node_controls)
-		process_controls(node_controls, ctx);
+		process_controls(node_controls, ctx, match);
 
 	return C_SUCCESS;
 }
@@ -1343,9 +1372,9 @@ static CResult process_device (const xmlNode *node_device, ParseContext *ctx)
  * Process a @c devices node.
  */
 static CResult process_devices (const xmlNode *node_devices, ParseContext *ctx)
-{
+{	
 	assert(node_devices);
-
+	
 	// Process all <device> nodes
 	xmlNode *node_device = xml_get_first_child_by_name(node_devices, "device");
 	while(node_device) {
@@ -1492,13 +1521,13 @@ static CResult process_meta (const xmlNode *node_meta, ParseContext *ctx)
 /**
  * Process an XML document tree representing a dynamic controls configuration file.
  */
-static CResult process_dynctrl_doc (xmlDoc *xml_doc, ParseContext *ctx)
+static CResult process_dynctrl_doc (ParseContext *ctx)
 {
 	CResult ret = C_SUCCESS;
-	if(!xml_doc)
-		return C_INVALID_ARG;
+	//if(!xml_doc)
+	//	return C_INVALID_ARG;
 
-	xmlNode *node_root = xmlDocGetRootElement(xml_doc);
+	xmlNode *node_root = xmlDocGetRootElement(ctx->xml_doc);
 	assert(node_root);
 	ctx->pass++;	// We start at pass 1 ...
 
@@ -1573,8 +1602,41 @@ static CResult device_supports_dynctrl(ParseContext *ctx)
 		/* Success: Assume not supported */
 		ret = C_NOT_IMPLEMENTED;
 	}
+#else
+	struct uvc_xu_control_mapping xu_control = {
+		.id = V4L2_CID_BRIGHTNESS,
+		.name = "Brightness",
+		.entity = /* UVC_GUID_UVC_PROCESSING */ { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01 },
+		.selector	= /* PU_BRIGHTNESS_CONTROL */ 0x02,
+		.size = 0,
+		.offset = 0,
+		.v4l2_type = V4L2_CTRL_TYPE_INTEGER,
+		.data_type = UVC_CTRL_DATA_TYPE_SIGNED,
+		.menu_info = NULL,
+		.menu_count = 0,
+		.reserved = {0,0,0,0},
+	};
+	
+	int v4l2_ret = ioctl(ctx->v4l2_handle, UVCIOC_CTRL_MAP, &xu_control);
+	if(v4l2_ret == -1) {
+		if(errno == EPERM) {
+			/* User is not root (newer drivers require root permissions) */
+			ret = C_CANNOT_WRITE;
+		}
+		else if(errno == ENOENT) {
+			/* Driver supports dynamic controls */
+		}
+		else {
+			printf("UVCIOC_CTRL_MAP returned errno(%i)\n", errno);
+			/* Unexpected error: Assume not supported */
+			ret = C_NOT_IMPLEMENTED;
+		}
+	}
+	else {
+		/* Success: Assume not supported */
+		ret = C_NOT_IMPLEMENTED;
+	}
 #endif
-
 	return ret;
 }
 
@@ -1582,7 +1644,6 @@ static CResult device_supports_dynctrl(ParseContext *ctx)
 /** 
  * Adds controls and control mappings contained in the given XML tree to the UVC driver.
  *
- * @param xml_doc	XML document tree corresponding to the dynctrl format
  * @param ctx		current parse context
  *
  * @return
@@ -1590,7 +1651,7 @@ static CResult device_supports_dynctrl(ParseContext *ctx)
  * 		- #C_CANNOT_WRITE if the user does not have permissions to add the mappings
  * 		- #C_SUCCESS if adding the controls and control mappings was successful
  */
-static CResult add_control_mappings (xmlDoc *xml_doc, ParseContext *ctx)
+static CResult add_control_mappings (ParseContext *ctx)
 {
 	CResult ret = C_SUCCESS;
 
@@ -1609,9 +1670,35 @@ static CResult add_control_mappings (xmlDoc *xml_doc, ParseContext *ctx)
 	if(ret) goto done;
 
 	// Process the contained control mappings
-	ret = process_dynctrl_doc(xml_doc, ctx);
+	ret = process_dynctrl_doc(ctx);
 
 done:
+	if (ret != C_SUCCESS) {
+		if(ret == C_NOT_IMPLEMENTED) {
+			add_error(ctx,
+				"device '%s' skipped because the driver '%s' behind it does not seem "
+				"to support dynamic controls.",
+				ctx->device->shortName, ctx->device->driver
+			);
+		}
+		else if(ret == C_CANNOT_WRITE) {
+			add_error(ctx,
+				"device '%s' skipped because you do not have the right permissions. "
+				"Newer driver versions require root permissions.",
+				ctx->device->shortName
+			);
+		}
+		else {
+			char *error = c_get_handle_error_text(ctx->handle, ret);
+			assert(error);
+			add_error(ctx,
+				"device '%s' was not processed successfully: %s. (Code: %d)",
+				ctx->device->shortName, error, ret
+			);
+			free(error);
+		}
+	}
+
 	// Close the device handle
 	if(ctx && ctx->v4l2_handle) {
 		close(ctx->v4l2_handle);
@@ -1621,7 +1708,88 @@ done:
 	return ret;
 }
 
+/**
+ * Cleans up all resources held by the passed in ctx
+ *
+ * @param ctx		current parse context
+ */
+static void destroy_context (ParseContext *ctx)
+{
+	if(!ctx) return;
 
+	// Close the conversion descriptor
+	if(ctx->cd && ctx->cd != (iconv_t)-1)
+		iconv_close(ctx->cd);
+
+	if(ctx->xml_doc)
+		xmlFreeDoc(ctx->xml_doc);
+
+	// Free the ParseContext.constants list
+	Constant *celem = ctx->constants;
+	while(celem) {
+		Constant *next = celem->next;
+		free(celem->name);
+		free(celem);
+		celem = next;
+	}
+
+	// Free the ParseContext.controls list
+	UVCXUControl *elem = ctx->controls;
+	while(elem) {
+		UVCXUControl *next = elem->next;
+		xmlFree(elem->id);
+		free(elem);
+		elem = next;
+	}
+
+	free(ctx);
+}
+
+/**
+ * Create a parsing context using the specified XML config-file
+ *
+ * @param file_name	name (with an optional path) of the file to be parsed
+ * @param info		structure to pass operation flags and retrieve status information.
+ * 				Can be NULL.
+ * @param ctx		address of a pointer to store the created context
+ *
+ * @return
+ * 		- C_NO_MEMORY if a buffer or structure could not be allocated
+ * 		- C_PARSE_ERROR if the XML file is malformed
+ * 		- C_SUCCESS if parsing was successful
+ */
+static CResult create_context (const char *file_name, CDynctrlInfo *info,
+			       ParseContext **ctx_ret)
+{
+	ParseContext *ctx;
+	CResult ret;
+
+	*ctx_ret = NULL;
+
+	// Allocate memory for the parsing context
+	ctx = (ParseContext *)malloc(sizeof(ParseContext));
+	if(!ctx) return C_NO_MEMORY;
+
+	memset(ctx, 0, sizeof(*ctx));
+	ctx->info = info;
+
+	// Parse the dynctrl configuration file
+	ret = parse_dynctrl_file(file_name, ctx);
+	if(ret) {
+		destroy_context(ctx);
+		return ret;
+	}
+
+	// Allocate a conversion descriptor
+	ctx->cd = iconv_open("ASCII", "UTF-8");
+	if(ctx->cd == (iconv_t)-1) {
+		destroy_context(ctx);
+		return C_NO_MEMORY;
+	}
+
+	*ctx_ret = ctx;
+	return C_SUCCESS;
+}
 
 /*
  * API
@@ -1656,7 +1824,6 @@ CResult c_add_control_mappings_from_file (const char *file_name, CDynctrlInfo *i
 	CResult ret = C_SUCCESS;
 	CDevice *devices = NULL;
 	ParseContext *ctx = NULL;
-	xmlDoc *xml_doc = NULL;
 
 	if(!initialized)
 		return C_INIT_ERROR;
@@ -1679,23 +1846,9 @@ CResult c_add_control_mappings_from_file (const char *file_name, CDynctrlInfo *i
 	ret = c_enum_devices(devices, &size, &device_count);
 	if(ret) goto done;
 
-	// Allocate memory for the parsing context
-	ctx = (ParseContext *)malloc(sizeof(ParseContext));
-	if(!ctx) {
-		ret = C_NO_MEMORY;
-		goto done;
-	}
-	memset(ctx, 0, sizeof(*ctx));
-	ctx->info = info;
-
-	// Parse the dynctrl configuration file
-	ret = parse_dynctrl_file(file_name, &xml_doc, ctx);
+	ret = create_context(file_name, info, &ctx);
 	if(ret) goto done;
 
-	// Allocate a conversion descriptor
-	ctx->cd = iconv_open("ASCII", "UTF-8");
-	assert(ctx->cd != (iconv_t)-1);
-
 	// Loop through the devices and check which ones have a supported uvcvideo driver behind them
 	int i, successful_devices = 0;
 	for(i = 0; i < device_count; i++) {
@@ -1719,76 +1872,87 @@ CResult c_add_control_mappings_from_file (const char *file_name, CDynctrlInfo *i
 			);
 			continue;
 		}
-
+		ctx->device = device;
+		
 		// Add the parsed control mappings to this device
-		ret = add_control_mappings(xml_doc, ctx);
+		ret = add_control_mappings(ctx);
 		if(ret == C_SUCCESS) {
 			successful_devices++;
 		}
-		else if(ret == C_NOT_IMPLEMENTED) {
-			add_error(ctx,
-				"device '%s' skipped because the driver '%s' behind it does not seem "
-				"to support dynamic controls.",
-				device->shortName, device->driver
-			);
-		}
-		else if(ret == C_CANNOT_WRITE) {
-			add_error(ctx,
-				"device '%s' skipped because you do not have the right permissions. "
-				"Newer driver versions require root permissions.",
-				device->shortName
-			);
-		}
-		else {
-			char *error = c_get_handle_error_text(ctx->handle, ret);
-			assert(error);
-			add_error(ctx,
-				"device '%s' was not processed successfully: %s. (Code: %d)",
-				device->shortName, error, ret
-			);
-			free(error);
-		}
 
 		// Close the device handle
 		c_close_device(ctx->handle);
 		ctx->handle = 0;
+		ctx->device = NULL;
 	}
 	if(successful_devices == 0)
 		ret = C_INVALID_DEVICE;
 
 done:
-	// Close the conversion descriptor
-	if(ctx && ctx->cd && ctx->cd != (iconv_t)-1)
-		iconv_close(ctx->cd);
-
-	// Clean up
-	if(xml_doc) xmlFreeDoc(xml_doc);
-	if(ctx) {
-		// Free the ParseContext.constants list
-		Constant *celem = ctx->constants;
-		while(celem) {
-			Constant *next = celem->next;
-			free(celem->name);
-			free(celem);
-			celem = next;
-		}
-
-		// Free the ParseContext.controls list
-		UVCXUControl *elem = ctx->controls;
-		while(elem) {
-			UVCXUControl *next = elem->next;
-			xmlFree(elem->id);
-			free(elem);
-			elem = next;
-		}
-
-		free(ctx);
-	}
+	destroy_context(ctx);
 	if(devices) free(devices);
 
 	return ret;
 }
 
+/**
+ * Parses a dynamic controls configuration file and adds the contained controls and
+ * control mappings to the UVC device pointed to by the passed in handle.
+ *
+ * Notes:
+ * - If the @a info parameter is not NULL the caller must free the info->messages field
+ *   if it is not NULL.
+ * - Note that this function is not thread-safe.
+ *
+ * @param handle		handle to the device to add control mappings to
+ * @param file_name		name of the controls configuration file
+ * @param info			structure to pass operation flags and retrieve status information.
+ * 						Can be NULL.
+ *
+ * @return
+ * 		- #C_INIT_ERROR if the library has not been initialized
+ * 		- #C_NO_MEMORY if memory could not be allocated
+ * 		- #C_SUCCESS if the parsing was successful and no fatal error occurred
+ * 		- #C_NOT_IMPLEMENTED if libwebcam was compiled with dynctrl support disabled
+ */
+CResult c_add_control_mappings (CHandle handle, const char *file_name,
+				CDynctrlInfo *info)
+{
+	CResult ret;
+	CDevice *device = NULL;
+	ParseContext *ctx = NULL;
+	unsigned int size = 0;
+
+	if(!initialized)
+		return C_INIT_ERROR;
+	if(!handle)
+		return C_INVALID_ARG;
+	if(!file_name)
+		return C_INVALID_ARG;
+
+	ret = c_get_device_info(handle, NULL, device, &size);
+	if(ret != C_BUFFER_TOO_SMALL) {
+		// Something bad has happened, so bail out
+		return ret;
+	}
+
+	device = (CDevice *)malloc(size);
+	ret = c_get_device_info(handle, NULL, device, &size);
+	if(ret) goto done;
+
+	ret = create_context(file_name, info, &ctx);
+	if(ret) goto done;
+
+	ctx->handle = handle;
+	ctx->device = device;
+	ret = add_control_mappings(ctx);
+
+done:
+	destroy_context(ctx);
+	free(device);
+
+	return ret;
+}
 
 #else
 
@@ -1798,5 +1962,10 @@ CResult c_add_control_mappings_from_file (const char *file_name, CDynctrlInfo *i
 	return C_NOT_IMPLEMENTED;
 }
 
+CResult c_add_control_mappings (CHandle handle, const char *file_name,
+				CDynctrlInfo *info)
+{
+	return C_NOT_IMPLEMENTED;
+}
 
 #endif
