@@ -2,7 +2,7 @@
  * uvcdynctrl - Manage dynamic controls in uvcvideo
  *
  *
- * Copyright (c) 2006-2008 Logitech.
+ * Copyright (c) 2006-2010 Logitech.
  *
  * This file is part of uvcdynctrl.
  * 
@@ -23,7 +23,22 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
+#include <dirent.h>
+#include <fnmatch.h>
+#include <ctype.h>
+#include <assert.h>
+#include <unistd.h>
+#include <errno.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <sys/ioctl.h>
+#include <fcntl.h>
+#include <linux/types.h>
 
+#define __user
+#include <linux/media.h>
+
+#include "config.h"
 #include "cmdline.h"
 #include "controls.h"
 
@@ -71,6 +86,7 @@ const char *get_control_type (CControlType type)
 		case CC_TYPE_BYTE:		return "Byte";
 		case CC_TYPE_WORD:		return "Word";
 		case CC_TYPE_DWORD:		return "Dword";
+		case CC_TYPE_BUTTON:	return "Button";
 		default:
 			return "<Unknown>";
 	}
@@ -186,6 +202,33 @@ print_device (CDevice *device)
 	}
 }
 
+static int file_exists (char * fileName)
+{
+	int result = access (fileName, F_OK); 
+   
+	/* File found */
+	if ( result == 0 )
+    {
+    	return 1;
+    }
+    return 0;   
+}
+
+static CResult
+save_controls (CHandle hDevice, const char* filename)
+{
+    int ret = C_SUCCESS;
+    ret = c_save_controls (hDevice, filename);
+    return ret;
+}
+
+static CResult
+load_controls (CHandle hDevice, const char* filename)
+{
+    int ret = C_SUCCESS;
+	ret = c_load_controls (hDevice, filename);
+    return ret;
+}
 
 static CResult
 list_controls (CHandle hDevice)
@@ -416,6 +459,121 @@ done:
 	return ret;
 }
 
+static CResult
+list_entities (CDevice *device)
+{
+	assert(device != NULL);
+
+	int device_index = -1;
+	if(sscanf(device->shortName, "video%u", &device_index) != 1)
+		return C_INVALID_ARG;
+
+	char media_path[32] = { 0 };
+	sprintf(media_path, "/dev/media%u", device_index);
+	//check for device node
+	if(file_exists(media_path))
+		printf("    Media controller device: %s\n", media_path);
+	else
+	{
+		printf("    Media controller device %s doesn't exist\n", media_path);
+		return C_INVALID_DEVICE;
+	}
+	
+	int mcdev = open(media_path, O_RDWR);
+	if(mcdev == -1) {
+		printf("ERROR: Unable to open media controller device '%s': %s (Error: %d)\n", media_path, strerror(errno), errno);
+		return C_INVALID_DEVICE;
+	}
+
+	int r = 0;
+	struct media_entity_desc entity;
+	memset(&entity, 0, sizeof(entity));
+	entity.id = 1;
+	while((r = ioctl(mcdev, MEDIA_IOC_ENUM_ENTITIES, &entity)) == 0) {
+		printf(
+			"    Entity %u: %s. Type: %u, Revision: %u, Flags: %u, Group-id: %u, Pads: %u, Links: %u\n",
+			entity.id, entity.name, entity.type, entity.revision, entity.flags, entity.group_id, entity.pads, entity.links
+		);
+
+		switch(entity.type & MEDIA_ENT_TYPE_MASK) {
+			case MEDIA_ENT_T_DEVNODE:
+				printf("      Device node\n");
+			
+				switch(entity.type & MEDIA_ENT_SUBTYPE_MASK) {
+					case MEDIA_ENT_T_DEVNODE_V4L:
+						printf("      Node: v4l: { major: %u, minor: %u }\n", entity.v4l.major, entity.v4l.minor);
+						break;
+					case MEDIA_ENT_T_DEVNODE_FB:
+						printf("      Node: fb: { major: %u, minor: %u }\n", entity.fb.major, entity.fb.minor);
+						break;
+					case MEDIA_ENT_T_DEVNODE_ALSA:
+						printf("      Node: alsa: {card: %u, device: %u, subdevice: %u}\n", entity.alsa.card, entity.alsa.device, entity.alsa.subdevice);
+						break;
+					case MEDIA_ENT_T_DEVNODE_DVB:
+						printf("      Node: dvb: %u\n", entity.dvb);
+						break;
+				}
+				break;
+				
+			case MEDIA_ENT_T_V4L2_SUBDEV:
+				printf("      Subdevice: ");
+			
+				switch(entity.type & MEDIA_ENT_SUBTYPE_MASK) {
+					case MEDIA_ENT_T_V4L2_SUBDEV_SENSOR:
+						printf(" Sensor\n");
+						break;
+					case MEDIA_ENT_T_V4L2_SUBDEV_FLASH:
+						printf(" Flash\n");
+						break;
+					case MEDIA_ENT_T_V4L2_SUBDEV_LENS:
+						printf(" Lens\n");
+						break;
+				}
+				break;
+		}
+
+		struct media_links_enum links;
+		memset(&links, 0, sizeof(links));
+		links.entity = entity.id;
+		links.pads = (struct media_pad_desc *)calloc(entity.pads, sizeof(struct media_pad_desc));
+		links.links = (struct media_link_desc *)calloc(entity.links, sizeof(struct media_link_desc));
+
+		r = ioctl(mcdev, MEDIA_IOC_ENUM_LINKS, &links);
+		if(r == 0) {
+			for(int i = 0; i < entity.pads; i++) {
+				struct media_pad_desc *pelem = &links.pads[i];
+				printf(
+					"      Entity: %u, Pad %u, Flags: %u\n",
+					pelem->entity, pelem->index, pelem->flags
+				);
+			}
+
+			for(int i = 0; i < entity.links; i++) {
+				struct media_link_desc *lelem = &links.links[i];
+				printf(
+					"      Out link: Source pad { Entity: %u, Index: %u, Flags: %u } => "
+					"Sink pad { Entity: %u, Index: %u, Flags: %u }\n",
+					lelem->source.entity, lelem->source.index, lelem->source.flags,
+					lelem->sink.entity, lelem->sink.index, lelem->sink.flags
+				);
+			}
+		}
+		else {
+			printf("ERROR: Unable to enumerate links for entity %u: %s (Error: %d)\n", entity.id, strerror(errno), errno);
+		}
+
+		free(links.links);
+		links.links = NULL;
+		free(links.pads);
+		links.pads = NULL;
+
+		entity.id++;
+	}
+
+	close(mcdev);
+	return C_SUCCESS;
+}
+
 
 static CResult
 list_devices ()
@@ -453,6 +611,12 @@ list_devices ()
 	for(int i = 0; i < count; i++) {
 		CDevice *device = &devices[i];
 		print_device(device);
+
+		ret = list_entities(device);
+		if(ret != C_SUCCESS && ret != C_INVALID_ARG) {
+			print_error("Unable to list device entities", ret);
+			ret = C_SUCCESS;
+		}
 	}
 
 done:
@@ -462,17 +626,99 @@ done:
 	return ret;
 }
 
+/*
+ * join path strings
+ */
+static char* 
+path_cat (const char * str1, char * str2) 
+{ 
+	size_t str1_len = strlen (str1); 
+	size_t str2_len = strlen (str2);
+	int nc;
+	
+	if ((str1[str1_len] != '/') && (str2[0] != '/')) nc = 2;
+	else nc = 1;
+	 
+	char * result; 
+	result = malloc ((str1_len + str2_len + nc) * sizeof * result); 
+	strcpy (result, str1); 
+	int i, j;
+	i = str1_len;
+	if (nc > 1) result[i] = '/';
+	i++; 
+	for (j = 0; ((i <(str1_len + str2_len + (nc-1))) && (j <str2_len)); j++) 
+	{ 
+		result [i] = str2 [j];
+		i++; 
+	}
+	result [str1_len + str2_len + (nc-1)] = '\0'; 
+	return result; 
+}
+
+
+/*
+ * get a xml filename NULL terminated list from data dir 
+ */
+static char**
+get_filename (const char *dir_path, const char *vid)
+{
+	struct dirent * dp;
+	struct dirent * sdp;
+	char** file_list;
+	int nf=1;
+	DIR * dir = opendir(dir_path);
+	
+	file_list = calloc(1, sizeof(file_list));
+	file_list[0] = NULL;
+	printf ( "checking dir: %s \n", dir_path);
+	while ((dp = readdir(dir)) != NULL) 
+	{
+		if((dp->d_type == DT_DIR) && (fnmatch("[[:xdigit:]][[:xdigit:]][[:xdigit:]][[:xdigit:]]", dp->d_name, 0) == 0))
+		{
+			if( strcasecmp(vid, dp->d_name) != 0)
+			{
+				/*doesn't match - clean up and move to the next entry*/
+				continue;
+			}
+			
+			char *tmp = path_cat (dir_path, dp->d_name);
+			printf("found dir: %s \n", dp->d_name);
+			DIR * subdir = opendir(tmp);
+			while ((sdp = readdir(subdir)) != NULL) 
+			{
+				if( fnmatch("*.xml", sdp->d_name, 0) == 0 )
+				{
+					file_list[nf-1] = path_cat (tmp, sdp->d_name);
+					printf("found: %s \n", file_list[nf-1]);
+					nf++;
+					file_list = realloc(file_list,nf*sizeof(file_list));
+					file_list[nf-1] = NULL;   
+				} 
+			}
+			closedir(subdir);
+			free (tmp);
+			tmp = NULL;
+		}
+	} 
+	if(!file_list[0]) printf("Could not find %s entry\n", vid);
+	closedir (dir);
+	return(file_list);
+}
 
 static CResult
-add_control_mappings(const char *filename)
+add_control_mappings(CHandle hDevice, const char *filename)
 {
+	CResult res;
 	CDynctrlInfo info = { 0 };
 	info.flags = CD_REPORT_ERRORS;
 	if(HAS_VERBOSE())
 		info.flags |= CD_RETRIEVE_META_INFO;
 
 	printf("Importing dynamic controls from file %s.\n", filename);
-	CResult res = c_add_control_mappings_from_file(filename, &info);
+	if (hDevice)
+		res = c_add_control_mappings(hDevice, filename, &info);
+	else
+		res = c_add_control_mappings_from_file(filename, &info);
 	if(res)
 		print_error("Unable to import dynamic controls", res);
 
@@ -561,6 +807,16 @@ main (int argc, char **argv)
 
 	res = c_init();
 	if(res) goto done;
+	
+	// Open the device
+	if (!args_info.list_given && (!args_info.import_given || args_info.device_given)) {
+		handle = c_open_device(args_info.device_arg);
+		if(!handle) {
+			print_error("Unable to open device", -1);
+			res = C_INVALID_DEVICE;
+			goto done;
+		}
+	}
 
 	// List devices
 	if(args_info.list_given) {
@@ -569,15 +825,93 @@ main (int argc, char **argv)
 	}
 	// Import dynamic controls from XML file
 	else if(args_info.import_given) {
-		res = add_control_mappings(args_info.import_arg);
+		res = add_control_mappings(handle, args_info.import_arg);
 		goto done;
 	}
-
-	// Open the device
-	handle = c_open_device(args_info.device_arg);
-	if(!handle) {
-		print_error("Unable to open device", -1);
-		res = C_INVALID_DEVICE;
+	// Import dynamic controls from XML files at default location
+	if(args_info.addctrl_given) {
+		// list all xml files at default data/vid dir
+		int nf=0;
+		char vid[5];
+		char pid[5];
+		short pid_set = 0;
+		if(fnmatch("[[:xdigit:]][[:xdigit:]][[:xdigit:]][[:xdigit:]]:[[:xdigit:]][[:xdigit:]][[:xdigit:]][[:xdigit:]]", args_info.addctrl_arg, 0))
+		{
+			if(fnmatch("[[:xdigit:]][[:xdigit:]][[:xdigit:]][[:xdigit:]]", args_info.addctrl_arg, 0))
+			{
+				printf("%s invalid: set at least a valid vid value, :pid is optional\n", args_info.addctrl_arg);
+				goto done;
+			}
+			else
+			{
+				/*extract vid and reset pid*/
+				int c = 0;
+				for (c = 0; c < 4; c++)
+				{
+					vid[c] = args_info.addctrl_arg[c];
+					pid[c] = 0;
+				}
+				vid[4] = '\0';
+				pid[4] = '\0';
+			}
+		}
+		else 
+		{
+			/*extract vid and pid*/
+			int c = 0;
+			for (c = 0; c < 4; c++)
+			{
+				vid[c] = args_info.addctrl_arg[c];
+				pid[c] = args_info.addctrl_arg[c+5];
+			}
+			vid[4] = '\0';
+			pid[4] = '\0';
+			pid_set = 1; /*flag pid.xml check*/
+			//printf("vid:%s pid:%s\n", vid, pid);
+		}
+		
+		/* get xml file list from DATA_DIR/vid/ */ 
+		char **xml_files = get_filename (DATA_DIR, vid);
+ 
+		/*check for pid.xml*/
+		char fname[9];
+		strcpy(fname, pid);
+		strcat(fname,".xml");
+		if(pid_set)
+		{
+			pid_set = 0; /*reset*/
+			nf=0;
+			while (xml_files[nf] != NULL)
+			{
+				if ( strcasecmp(fname, xml_files[nf]) == 0)
+					pid_set = 1; /*file exists so flag it*/
+				nf++;
+			}
+		}
+		
+		/*parse xml files*/
+		nf = 0;
+		while (xml_files[nf] != NULL)
+		{
+			/* if pid was set and pid.xml exists parse it*/
+			if(pid_set)
+			{
+				if ((strcasecmp(fname, xml_files[nf]) == 0))
+				{
+					printf ( "Parsing: %s \n", xml_files[nf]);
+					res = add_control_mappings(handle, xml_files[nf]);
+				}
+			}
+			else /* parse all xml files inside vid dir */
+			{
+				printf ( "Parsing: %s \n", xml_files[nf]);
+				res = add_control_mappings(handle, xml_files[nf]);
+			}
+			free(xml_files[nf]);
+			xml_files[nf]=NULL;
+			nf++;
+		}
+		free(xml_files);
 		goto done;
 	}
 
@@ -641,6 +975,12 @@ main (int argc, char **argv)
 			goto done;
 		}
 	}
+	else if(args_info.save_ctrl_given) {
+		res = save_controls( handle, args_info.save_ctrl_arg);
+	}
+	else if(args_info.load_ctrl_given) {
+		res = load_controls( handle, args_info.load_ctrl_arg);
+	}
 
 	// Clean up
 done:
